open Util

open Kernel

open Format

let map_add x s a var_map =
  let sgn_map = Opt.var_map_get x var_map in
  let list = Opt.sgn_list_get s sgn_map in
  let sgn_map = Opt.sgn_list_add s (a :: list) sgn_map in
  Opt.var_map_add x sgn_map var_map

let map_is_single var_map =
  try
    let found = ref false in
    let find_sgn _ list =
      match list with
        | [_] when not !found ->
            found := true
        | _ ->
            raise Exit in
    let find_var _ sgn_map =
      Sgn.Map.iter find_sgn sgn_map in
    Var.Map.iter find_var var_map;
    !found
  with Exit -> false

let restore_enabled = ref true

let disable_restore () =
  restore_enabled := false

let () =
  Cmdline.register ~long:"disable-restore"
    (Cmdline.Zero disable_restore)
    ~help:"disables restore optimization"

module Indexed = struct
  type t = {
    vars: Var.Set.t;
    restores: (Var.t * int) list;
    tells: (Pos.range * Value.t array) list Sgn.Map.t Var.Map.t;
    calls: (Pos.range * Opt.Call.t) list;
    asks: (Pos.range * Var.t array * Var.t * int * t) list Sgn.Map.t Var.Map.t;
    bool_calls: (Pos.range * Sgn.t * Value.t array * t * t) list;
  }

  let add_tell asks depth (restores, list) (pos, a) =
    try
      let get_var v =
        match v with
          | Value.Var v -> v
          | _ -> raise Exit in
      let a = Array.map get_var a in
      let (r, depth') = List.assoc a asks in
      ((r, depth - depth' - 1) :: restores, list)
    with Exit | Not_found ->
      (restores, (pos, a) :: list)

  let add_tell_sgn asks depth s list (restores, sgn_map) =
    try
      let asks' = Sgn.Map.find s asks in
      let (restores, list) =
        List.fold_left (add_tell asks' depth) (restores, []) list in
      (restores, Opt.sgn_list_add s list sgn_map)
    with Not_found ->
      (restores, Sgn.Map.add s list sgn_map)

  let add_tell_var asks depth x sgn_map (restores, var_map) =
    try
      let asks' = Var.Map.find x asks in
      let (restores, sgn_map) =
        Sgn.Map.fold (add_tell_sgn asks' depth)
          sgn_map (restores, Sgn.Map.empty) in
      (restores, Opt.var_map_add x sgn_map var_map)
    with Not_found ->
      (restores, Var.Map.add x sgn_map var_map)

  let rec of_opt' counter parent_asks depth opt =
    let succ_asks =
      if Var.Map.is_empty opt.Opt.tells && opt.Opt.bool_calls = []
        && map_is_single opt.Opt.asks then
          parent_asks
      else
        Var.Map.empty in
    let (restores, tells) =
      if !restore_enabled then
        let (restores, tells) =
          Var.Map.fold (add_tell_var parent_asks depth)
            opt.Opt.tells ([], Var.Map.empty) in
        let restores = List.sort (fun (_, d) (_, d') -> compare d d') restores in
        (restores, tells)
      else
        ([], opt.Opt.tells) in
    let index_ask x s (p, a, opt) =
      let index = !counter in
      let v = Var.fresh () in
      counter := succ index;
      let parent_asks = map_add x s (a, (v, depth)) succ_asks in 
      let indexed = of_opt' counter parent_asks (depth + 1) opt in
      (p, a, v, index, indexed) in
    let asks =
      Var.Map.mapi
        (fun x m -> Sgn.Map.mapi
           (fun s l -> List.map (index_ask x s) l) m) opt.Opt.asks in
    let index_bool_call (p, s, a, t, f) =
      let t = of_opt' counter parent_asks depth t in
      let f = of_opt' counter parent_asks depth f in
      (p, s, a, t, f) in
    let bool_calls = List.map index_bool_call opt.Opt.bool_calls in
    { vars = opt.Opt.vars;
      restores = restores;
      tells = tells;
      calls = opt.Opt.calls;
      asks = asks;
      bool_calls = bool_calls;
    }

  let of_opt opt =
    let counter = ref 1 in
    let indexed = of_opt' counter Var.Map.empty 0 opt in
    (indexed, !counter)
end

let tell_calls_enabled = ref true

let disable_tell_calls () =
  tell_calls_enabled := false

let () =
  Cmdline.register ~long:"disable-tell-calls"
    (Cmdline.Zero disable_tell_calls)
    ~help:"disables tell calls optimization"

module Reduced = struct
  type t = {
    vars: Var.Set.t;
    restores: (Var.t * int) list;
    tells: (Pos.range * Value.t array) list Sgn.Map.t Var.Map.t;
    tell_calls: (Pos.range * Value.t array * int) list;
    calls: (Pos.range * Opt.Call.t) list;
    asks: (Pos.range * Var.t array * Var.t * int * t) list Sgn.Map.t Var.Map.t;
    bool_calls: (Pos.range * Sgn.t * Value.t array * t * t) list;
    safe_asks_set: Int.Set.t;
  }

  let rec no_sub_asks_rec x s indexed =
    Opt.map_iter (fun _ _ (_, _, _, _, code) -> no_asks x s code)
      indexed.Indexed.asks;
    List.iter (fun (_, _, _, t, f) -> no_asks x s t; no_asks x s f)
      indexed.Indexed.bool_calls
  and no_asks x s indexed =
    try
      let sgn_map = Var.Map.find x indexed.Indexed.asks in
      let _ = Sgn.Map.find s sgn_map in
      raise Exit
    with Not_found ->
      no_sub_asks_rec x s indexed

  let no_sub_asks x s indexed =
    try
      no_sub_asks_rec x s indexed;
      true
    with Exit -> false

  let add_safe_asks_sgn indexed x s list (safe_asks, set) =
    match list with
      | [(p, a, v, id, code)] ->
          if no_sub_asks x s indexed then
            let safe_asks = Sgn.Map.add s (p, a, v, id, code) safe_asks in
            let set = Int.Set.add id set in
            (safe_asks, set)
          else
            (safe_asks, set)
      | _ -> (safe_asks, set)

  let add_safe_asks_var indexed unsafe_variables asks x (safe_asks, set) =
    if Var.Set.mem x unsafe_variables then
      (safe_asks, set)
    else
      try
        let sgn_map = Var.Map.find x asks in
        let safe_asks' = Opt.var_map_get x safe_asks in
        let (safe_asks', set) =
          Sgn.Map.fold (add_safe_asks_sgn indexed x) sgn_map
            (safe_asks', set) in
        let safe_asks = Opt.var_map_add x safe_asks' safe_asks in
        (safe_asks, set)
      with Not_found -> (safe_asks, set)

  let add_tell id tell_calls (p, a) = (p, a, id) :: tell_calls

  let add_tell_sgn safe_asks s list (tells, tell_calls) =
    try
      let (_, _, _, id, _) = Sgn.Map.find s safe_asks in
      let tell_calls = List.fold_left (add_tell id) tell_calls list in
      (tells, tell_calls)
    with Not_found ->
      (Sgn.Map.add s list tells, tell_calls)

  let add_tell_var safe_asks x sgn_map (tells, tell_calls) =
    try
      let safe_asks' = Var.Map.find x safe_asks in
      let (tells', tell_calls) =
        Sgn.Map.fold (add_tell_sgn safe_asks') sgn_map
          (Sgn.Map.empty, tell_calls) in
      (Opt.var_map_add x tells' tells, tell_calls)
    with Not_found ->
      (Var.Map.add x sgn_map tells, tell_calls)

  let rec of_indexed unsafe_variables safe_asks indexed =
    let asks = indexed.Indexed.asks in
    let (safe_asks, safe_asks_set) =
      if !tell_calls_enabled then
        Var.Set.fold
          (add_safe_asks_var indexed unsafe_variables asks)
          indexed.Indexed.vars (safe_asks, Int.Set.empty)
      else
        (safe_asks, Int.Set.empty) in
    let tells, tell_calls =
      Var.Map.fold (add_tell_var safe_asks) indexed.Indexed.tells
        (Var.Map.empty, []) in
    let rec_ask (p, a, v, id, code) =
      (p, a, v, id, of_indexed unsafe_variables safe_asks code) in
    let asks = Var.Map.map (Sgn.Map.map (List.map rec_ask)) asks in
    let rec_bool_call (p, s, a, t, f) =
      let t = of_indexed unsafe_variables safe_asks t in
      let f = of_indexed unsafe_variables safe_asks f in
      (p, s, a, t, f) in
    let bool_calls = List.map rec_bool_call indexed.Indexed.bool_calls in
    { vars = indexed.Indexed.vars;
      restores = indexed.Indexed.restores;
      tells = tells;
      tell_calls = tell_calls;
      calls = indexed.Indexed.calls;
      asks = asks;
      bool_calls = bool_calls;
      safe_asks_set = safe_asks_set;
    }
end

type indexed = {
  count: int;
  map: int Var.Map.t;
}

let index_var x indexed = {
  count = succ indexed.count;
  map = Var.Map.add x indexed.count indexed.map
}

let index_var_set_from indexed set =
  Var.Set.fold index_var set indexed

let indexed_empty = { count = 0; map = Var.Map.empty }

let index_var_set set =
  index_var_set_from indexed_empty set

let index_var_array a =
  let count = Array.length a in
  let rec init_map i map =
    if i < count then
      init_map (succ i) (Var.Map.add a.(i) i map)
    else
      map in
  let map = init_map 0 Var.Map.empty in
  { count = count; map = map }

type ('a, 'b) ctx = {
  local_vars: 'a;
  param_vars: 'a;
  env_vars: 'a;
  global_vars: 'a;
  ask_var: 'b;
}

type var =
  | Local of int
  | Param of int
  | Env of int
  | Global of int
  | Ask_var

type value =
  | Var of var
  | Term of value Term.t

let get_var var_tbl v =
  try Local (Var.Map.find v var_tbl.local_vars.map)
  with Not_found ->
    if var_tbl.ask_var = Some v then
      Ask_var
    else
      try Param (Var.Map.find v var_tbl.param_vars.map)
      with Not_found ->
        try Env (Var.Map.find v var_tbl.env_vars.map)
        with Not_found ->
          try Global (Var.Map.find v var_tbl.global_vars.map)
          with Not_found -> failwith ("get_var: " ^ v)

let rec get_value var_tbl v =
  match v with
    | Value.Var v -> Var (get_var var_tbl v)
    | Value.Term t -> Term (Term.map (get_value var_tbl) t)

module Env_tbl = struct
  type entry = {
    mutable env: Var.Set.t;
    mutable local: Var.Set.t;
    mutable params: Var.Set.t;
    mutable param_array: Var.t array;
    mutable super: int list;
    mutable ask_v: Var.t option;
  }

  let create_entry () = {
    env = Var.Set.empty;
    local = Var.Set.empty;
    params = Var.Set.empty;
    param_array = [| |];
    super = [];
    ask_v = None;
  }

  let rec add_vars env_tbl set index =
    let entry = env_tbl.(index) in
    let set = Var.Set.diff set entry.local in
    let set = Var.Set.diff set entry.params in
    let set = Var.Set.diff set entry.env in
    let set = Var.Set.diff set env_tbl.(0).local in
    if not (Var.Set.is_empty set) then
      begin
        entry.env <- Var.Set.union set entry.env;
        List.iter (add_vars env_tbl set) entry.super
      end

  let add_sub env_tbl sub parent =
    let entry = env_tbl.(sub) in
    add_vars env_tbl entry.env parent;
    entry.super <- parent :: entry.super

  let enum_vars_in_tell_var_map vars var_map =
    let add_tells x map vars =
      Opt.enum_vars_in_tell_sgn_map (Var.Set.add x vars) map in
    Var.Map.fold add_tells var_map vars

  let enum_vars_in_tell_call vars (_, a, _) =
    Array.fold_left Opt.enum_vars_in_value vars a

  let add_tell_call env_tbl index (_, a, index') =
    add_sub env_tbl index' index

  let add_ask_var x _ vars = Var.Set.add x vars

  let enum_vars_in_bool_ask vars (_, _, a, _, _) =
    Array.fold_left Opt.enum_vars_in_value vars a

  let rec make' env_tbl index reduced =
    let entry = env_tbl.(index) in
    entry.local <- Var.Set.union entry.local reduced.Reduced.vars;
    let vars = enum_vars_in_tell_var_map Var.Set.empty reduced.Reduced.tells in
    let vars =
      List.fold_left enum_vars_in_tell_call vars reduced.Reduced.tell_calls in
    let vars =
      List.fold_left Opt.enum_vars_in_call vars reduced.Reduced.calls in
    let vars = Var.Map.fold add_ask_var reduced.Reduced.asks vars in
    let vars =
      List.fold_left enum_vars_in_bool_ask vars reduced.Reduced.bool_calls in
    let vars =
      List.fold_left (fun vars (v, _) -> Var.Set.add v vars)
        vars reduced.Reduced.restores in
    add_vars env_tbl vars index;
    List.iter (add_tell_call env_tbl index) reduced.Reduced.tell_calls;
    let add_ask _ _ (_, a, v, index', reduced) =
      add_sub env_tbl index' index;
      let entry' = env_tbl.(index') in
      entry'.params <- Opt.set_of_var_array (Var.Set.singleton v) a;
      entry'.param_array <- a;
      entry'.ask_v <- Some v;
      make' env_tbl index' reduced in
    Opt.map_iter add_ask reduced.Reduced.asks;
    let add_bool_call (_, _, _, t, f) =
      make' env_tbl index t;
      make' env_tbl index f in
    List.iter add_bool_call reduced.Reduced.bool_calls    

  let make_env global_vars entry = {
    env_vars = index_var_set entry.env;
    local_vars = index_var_set entry.local;
    param_vars = index_var_array entry.param_array;
    global_vars = global_vars;
    ask_var = entry.ask_v;
  }

  let make reduced count old_global_vars =
    let env_tbl = Array.init count (fun _ -> create_entry ()) in
    make' env_tbl 0 reduced;
    let global_vars =
      index_var_set_from old_global_vars env_tbl.(0).local in
    let var_tbl = Array.map (make_env global_vars) env_tbl in
    var_tbl.(0) <-
      { var_tbl.(0) with
          local_vars = global_vars;
          env_vars = indexed_empty };
    var_tbl
end

type call =
  | Void of Sgn.t * value array
  | Value of Sgn.t * value array * int

type agent_call = {
  pos: Pos.range;
  env: var array;
  id: int;
}

type ask_tell = {
  s: Sgn.t;
  asks: agent_call list;
  tells: (Pos.range * value array) list;
}

type code = {
  init: int list;
  restores: (var * int) list;
  calls: (Pos.range * call) list;
  tell_calls: (value array * agent_call) list;
  ask_tells: (var * ask_tell list) list;
  bool_calls: (Pos.range * Sgn.t * value array * code * code) list;
}

let format_var fmt v =
  match v with
    | Local i -> fprintf fmt "local:%i" i
    | Param i -> fprintf fmt "param:%i" i
    | Env i -> fprintf fmt "env:%i" i
    | Global i -> fprintf fmt "global:%i" i
    | Ask_var -> fprintf fmt "ask_var"

let rec format_value fmt value =
  match value with
    | Var v -> format_var fmt v
    | Term t -> Term.format format_value fmt t

let format_call fmt call =
  match call with
    | Void (s, a) ->
        fprintf fmt "%s(%a)" s.Sgn.f (Array.format "," format_value) a
    | Value (s, a, r) ->
        fprintf fmt "%d <- %s(%a)"
          r s.Sgn.f (Array.format "," format_value) a

let format_agent_call fmt agent_call =
  fprintf fmt "([%a],%d)"
    (Array.format "," format_var) agent_call.env agent_call.id

let format_tell_call fmt (params, agent_call) =
  fprintf fmt "([%a],%a)"
    (Array.format "," format_value) params format_agent_call agent_call

let format_tell fmt (p, a) =
  fprintf fmt "(%a)" (Array.format "," format_value) a

let format_ask_tell fmt ask_tell =
  fprintf fmt "(s:%s,asks:[%a],tells:[%a])"
    ask_tell.s.Sgn.f
    (List.format "," format_agent_call) ask_tell.asks
    (List.format ";" format_tell) ask_tell.tells

let format_ask_tells fmt (x, list) =
  fprintf fmt "(%a,%a)"
    format_var x (List.format "," format_ask_tell) list

let rec format_code fmt code =
  fprintf fmt "{\
    init:(%a);\
    restores:(%a);\
    calls:%a;\
    tell_calls:%a;\
    ask_tells:%a;\
    bool_calls:%a}@."
    (List.format "," pp_print_int) code.init
    (List.format "," (fun fmt (v, _) -> format_var fmt v)) code.restores
    (List.format "," (fun fmt (_, call) -> format_call fmt call)) code.calls
    (List.format "," format_tell_call) code.tell_calls
    (List.format "," format_ask_tells) code.ask_tells
    (List.format "," format_bool_calls) code.bool_calls
and format_bool_calls fmt (_, s, a, t, f) =
  fprintf fmt "if %s(%a) {%a} else{%a}"
    s.Sgn.f (Array.format "," format_value) a format_code t format_code f

type t = {
  env_count: int;
  local_count: int;
  param_count: int;
  code: code;
}

let format_proc fmt t =
  fprintf fmt "[env:%d;local:%d;param:%d]%a"
    t.env_count t.local_count t.param_count format_code t.code

let format fmt am =
  for i = 0 to Array.length am - 1 do
    fprintf fmt "{%d:%a}" i format_proc am.(i)
  done;
  pp_print_flush fmt ()

let empty = {
  env_count = 0;
  local_count = 0;
  param_count = 0;
  code = {
    init = [];
    restores = [];
    calls = [];
    tell_calls = [];
    ask_tells = [];
    bool_calls = [];
  }
}

let make_void_call var_tbl c =
  let s = c.Opt.Call.Void.builtin.Opt.Builtin.Void.desc.Opt.Builtin.s in
  let args = Array.map (get_value var_tbl) c.Opt.Call.Void.args in
  Void (s, args)

let make_value_call var_tbl c vars =
  let s = c.Opt.Call.Value.builtin.Opt.Builtin.Value.desc.Opt.Builtin.s in
  let args = Array.map (get_value var_tbl) c.Opt.Call.Value.args in
  let result = c.Opt.Call.Value.result in
  let vars = Var.Set.remove result vars in
  let result =
    try Var.Map.find result var_tbl.local_vars.map
    with Not_found -> failwith "make_value_call" in
  (vars, Value (s, args, result))

let make_call var_tbl (vars, calls) (pos, c) =
  let (vars, c) =
    match c with
      | Opt.Call.Void c -> (vars, make_void_call var_tbl c)
      | Opt.Call.Value c -> make_value_call var_tbl c vars in
  (vars, (pos, c) :: calls)

let make_env var_tbls var_tbl id =
  let var_tbl' = var_tbls.(id) in  
  let env = Array.make var_tbl'.env_vars.count (Local 0) in
  let add_var x index = env.(index) <- get_var var_tbl x in
  Var.Map.iter add_var var_tbl'.env_vars.map;
  env

let make_tell_call var_tbls var_tbl (pos, a, id) =
  let a = Array.map (get_value var_tbl) a in
  let env = make_env var_tbls var_tbl id in
  let call = { pos = pos; env = env; id = id } in
  (a, call)

let set_of_sgn_map set map =
  Sgn.Map.fold (fun s _ set -> Sgn.Set.add s set) map set

let make_ask make_proc var_tbls var_tbl safe_asks asks
    (pos, a, v, id, reduced) =
  make_proc var_tbls id reduced;
(*  if Int.Set.mem id safe_asks then
    asks
  else*)
    begin
      let env = make_env var_tbls var_tbl id in
      { pos = pos; env = env; id = id } :: asks
    end
      
let make_tell var_tbl (pos, a) = (pos, Array.map (get_value var_tbl) a)

let make_ask_tells_sgn make_proc var_tbls var_tbl safe_asks asks tells s list =
  let asks = Opt.sgn_list_get s asks in
  let asks =
    List.fold_left (make_ask make_proc var_tbls var_tbl safe_asks) [] asks in
  let tells = Opt.sgn_list_get s tells in
  let tells = List.map (make_tell var_tbl) tells in
  if asks = [] && tells = [] then
    list
  else
    { s = s; asks = asks; tells = tells } :: list

let make_ask_tells_var make_proc var_tbls var_tbl reduced x sgn list =
  let asks = Opt.var_map_get x reduced.Reduced.asks in
  let tells = Opt.var_map_get x reduced.Reduced.tells in
  let safe_asks = reduced.Reduced.safe_asks_set in
  let list' =
    Sgn.Set.fold
      (make_ask_tells_sgn make_proc var_tbls var_tbl safe_asks asks tells)
      sgn [] in
  (get_var var_tbl x, list') :: list

let add_sgn_set x sgn_map map =
  let set =
    try Var.Map.find x map
    with Not_found -> Sgn.Set.empty in
  Var.Map.add x (set_of_sgn_map set sgn_map) map

let make_bool_call make_code var_tbl (p, s, a, t, f) =
  let a = Array.map (get_value var_tbl) a in
  let t = make_code var_tbl t in
  let f = make_code var_tbl f in
  (p, s, a, t, f)

let rec make_code make_proc var_tbls var_tbl reduced =
  let (init_vars, calls) =
    List.fold_left (make_call var_tbl) (reduced.Reduced.vars, [])
      reduced.Reduced.calls in
  let calls = List.rev calls in
  let add_init_var x list =
    let i =
      try Var.Map.find x var_tbl.local_vars.map
      with Not_found -> failwith "add_init_var" in
    i :: list in
  let init = Var.Set.fold add_init_var init_vars [] in
  let restores =
    List.map (fun (v, depth) -> (get_var var_tbl v, depth))
      reduced.Reduced.restores in
  let tell_calls =
    List.map (make_tell_call var_tbls var_tbl) reduced.Reduced.tell_calls in
  let all_sgn =
    Var.Map.map (set_of_sgn_map Sgn.Set.empty) reduced.Reduced.asks in
  let all_sgn = Var.Map.fold add_sgn_set reduced.Reduced.tells all_sgn in
  let ask_tells =
    Var.Map.fold (make_ask_tells_var make_proc var_tbls var_tbl reduced)
      all_sgn [] in
  let bool_calls =
    List.map (make_bool_call (make_code make_proc var_tbls) var_tbl)
      reduced.Reduced.bool_calls in
  { init = init;
    restores = restores;
    calls = calls;
    tell_calls = tell_calls;
    ask_tells = ask_tells;
    bool_calls = bool_calls;
  }

let rec make_proc proc_tbl var_tbls index reduced =
  let var_tbl = var_tbls.(index) in
  proc_tbl.(index) <- {
    env_count = var_tbl.env_vars.count;
    local_count = var_tbl.local_vars.count;
    param_count = var_tbl.param_vars.count;
    code = make_code (make_proc proc_tbl) var_tbls var_tbl reduced;
  }

let print_code = ref false

let enable_print_code () =
  print_code := true

let () =
  Cmdline.register ~long:"print-code" (Cmdline.Zero enable_print_code)
    ~help:"print the intermediary code on standard error before executing it"

let of_opt opt old_global_vars =
  if !print_code then
    begin
      Opt.format err_formatter opt;
      pp_print_newline err_formatter ();
    end;
  let unsafe_variables = Opt.enum_unsafe_variables opt in
  let (indexed, count) = Indexed.of_opt opt in
  let reduced = Reduced.of_indexed unsafe_variables Var.Map.empty indexed in
  let var_tbls = Env_tbl.make reduced count old_global_vars in
  let proc_tbl = Array.make count empty in
  make_proc proc_tbl var_tbls 0 reduced;
  if !print_code then
    format err_formatter proc_tbl;
  (proc_tbl, var_tbls.(0).global_vars)
