module Id = struct
  let name = "pa_silcc_builtins"
  let version = "0.0.1"
end

module Pa_silcc_builtins (Syntax: Camlp4.Sig.Camlp4Syntax) = struct
  open Camlp4.PreCast

  include Syntax

  type module_table = {
    modules: (string, module_table) Hashtbl.t;
    types: (string, Ast.expr * Ast.expr) Hashtbl.t;
  }

  let create_module_table () = {
    modules = Hashtbl.create 17;
    types = Hashtbl.create 17;
  }

  type name =
    | Name of string
    | Mod of name * string

  let pervasives_table = create_module_table ()

  let () =
    let _loc = Loc.mk "string" in
    Hashtbl.add pervasives_table.types "string"
      (<:expr< Interpret.string_inj >>, <:expr< Interpret.string_proj>>)

  let () =
    let _loc = Loc.mk "int" in
    Hashtbl.add pervasives_table.types "int"
      (<:expr< Interpret.int_inj >>, <:expr< Interpret.int_proj>>)

  let () =
    let _loc = Loc.mk "float" in
    Hashtbl.add pervasives_table.types "float"
      (<:expr< Interpret.float_ty.Data.Type.inj >>,
       <:expr< Interpret.float_ty.Data.Type.proj>>)

  let () =
    let _loc = Loc.mk "big_int" in
    Hashtbl.add pervasives_table.types "big_int"
      (<:expr< Interpret.big_int_ty.Data.Type.inj >>,
       <:expr< Interpret.big_int_ty.Data.Type.proj >>)

  let () =
    let _loc = Loc.mk "bool" in
    Hashtbl.add pervasives_table.types "bool"
      (<:expr< Interpret.bool_inj >>,
       <:expr< Interpret.bool_proj >>)

  let () =
    let _loc = Loc.mk "char" in
    Hashtbl.add pervasives_table.types "char"
      (<:expr< Interpret.char_inj >>,
       <:expr< Interpret.char_proj >>)

  let () =
    let _loc = Loc.mk "mem" in
    Hashtbl.add pervasives_table.types "mem"
      (<:expr< Interpret.mem_ty.inj >>, <:expr< Interpret.mem_ty.proj>>)

  let () =
    let _loc = Loc.mk "value" in
    Hashtbl.add pervasives_table.types "value"
      (<:expr< (fun i -> i) >>, <:expr< (fun i -> i) >>)

  let mk_full_name _loc ident name =
    match ident with
      | None -> name
      | Some ident -> Ast.IdAcc (_loc, ident, name)

  let rec translate_type_sum name_ty prefix ident sum =
    let _loc = Ast.loc_of_ctyp sum in
    match sum with
      | Ast.TyOr (_, left, right) ->
          let left = translate_type_sum name_ty prefix ident left in
          let right = translate_type_sum name_ty prefix ident right in
          <:str_item< $left$ $right$ >>
      | Ast.TyId (_, (Ast.IdUid (_, name) as name_uid)) ->
          let name =
            if name.[0] = ' ' then
              String.sub name 1 (String.length name - 1)
            else
              name in
          let name_proj = prefix ^ "make_" ^ name in
          let name_def = name_proj in
          let full_name = Ast.ExId (_loc, mk_full_name _loc ident name_uid) in
          <:str_item<
            let $lid:name_def$ event t =
              $lid:name_ty$.Data.Type.inj $full_name$

            let () =
              Interpret.decl_value_builtin
                (Kernel.Sgn.get $str:name_proj$ 1) $lid:name_def$
          >>
      | _ ->
          <:str_item< >>
          

  let translate_type_rhs name_ty prefix ident rhs =
    let _loc = Ast.loc_of_ctyp rhs in
    match rhs with
      | Ast.TySum (_, sum) ->
          translate_type_sum name_ty prefix ident sum
      | _ ->
          <:str_item< >>

  let rec translate_type_declaration prefix ident current_table ty =
    let _loc = Ast.loc_of_ctyp ty in
    match ty with
      | Ast.TyDcl (_, name, [], rhs, cs) ->
          let prefix_name = prefix ^ name in
          let name_ty = prefix_name ^ "_ty" in
          let pp_print_name = "pp_print_" ^ prefix_name in
          let msg = "<" ^ prefix_name ^ ">" in
          let inj = <:expr< $lid:name_ty$.Data.Type.inj >> in
          let proj = <:expr< $lid:name_ty$.Data.Type.proj >> in
          let name_lid = Ast.IdLid (_loc, name) in
          let full_name = Ast.TyId (_loc, mk_full_name _loc ident name_lid) in
          let defs = translate_type_rhs name_ty prefix ident rhs in 
          Hashtbl.add current_table.types name (inj, proj);
          <:str_item<
            let $lid:pp_print_name$ fmt context =
              Format.fprintf fmt $str:msg$

            let $lid:name_ty$ : $full_name$ Data.Type.ty =
              Data.new_type $lid:pp_print_name$;;

            $defs$
          >>
      | Ast.TyAnd (_, ty0, ty1) ->
          let t0 = translate_type_declaration prefix ident current_table ty0 in
          let t1 = translate_type_declaration prefix ident current_table ty1 in
          <:str_item< $t0$ $t1$ >>
      | _ ->
          <:str_item< >>

  let rec lookup_data_ty current_table id =
    match id with
      | Ast.IdLid (_, id) ->
          Hashtbl.find current_table.types id
      | Ast.IdAcc (_, Ast.IdUid (_, m), id) ->
          lookup_data_ty (Hashtbl.find current_table.modules m) id
      | _ ->
          failwith "get_data_ty"

  let get_data_ty open_tbl current_tbl id =
    try lookup_data_ty current_tbl id
    with Not_found ->
      let rec lookup_open tables =
        match tables with
          | [] -> failwith "get_data_ty"
          | table :: tables ->
              begin
                try lookup_data_ty table id
                with Not_found -> lookup_open tables     
              end in
      lookup_open open_tbl

  let rec get_tup ty =
    match ty with
      | Ast.TySta (_, ty0, ty1) ->
          ty0 :: get_tup ty1
      | _ -> [ty]

  let counter = ref 0

  let new_binder () =
    let binder = Format.sprintf "a%d" !counter in
    counter := succ !counter;
    binder

  let rec get_ty open_tbl current_tbl ty =
    match ty with
      | Ast.TyId (_loc, ty) ->
          let (inj, proj) = get_data_ty open_tbl current_tbl ty in
          _loc, inj, proj
      | Ast.TyApp (_loc, Ast.TyId (_, Ast.IdLid (_, "array")), ty) ->
          let (_, inj, proj) = get_ty open_tbl current_tbl ty in
          _loc,
          <:expr<
            fun e -> Interpret.array_ty.Data.Type.inj (Array.map $inj$ e)
          >>,
          <:expr<
            fun e -> Array.map $proj$ (Interpret.array_ty.Data.Type.proj e)
          >>
      | Ast.TyApp (_loc, Ast.TyId (_, Ast.IdLid (_, "list")), ty) ->
          let (_, inj, proj) = get_ty open_tbl current_tbl ty in
          _loc,
          <:expr<
            fun e -> Interpret.list_inj (List.map $inj$ e)
          >>,
          <:expr<
            fun e -> List.map $proj$ (Interpret.list_proj e)
          >>
      | Ast.TyTup (_loc, ty) ->
          let tup = get_tup ty in
          let inj_pat, inj_arr, proj_tup = mk_tup open_tbl current_tbl 0 tup in
          let inj_pat = Ast.PaTup (_loc, inj_pat)
          and inj_arr = Ast.ExArr (_loc, inj_arr)
          and proj_tup = Ast.ExTup (_loc, proj_tup) in
          _loc,
          <:expr< fun $inj_pat$ -> Interpret.array_ty.Data.Type.inj $inj_arr$ >>,
          <:expr<
            fun e ->
              let a = Interpret.array_ty.Data.Type.proj e in
              $proj_tup$
          >>
      | _ ->
          failwith "get_ty"
  and mk_tup open_tbl current_tbl index ty =
    match ty with
      | [] -> failwith "mk_tup"
      | [last] ->
          let (_loc, inj, proj) = get_ty open_tbl current_tbl last in
          let binder = new_binder () in
          <:patt< $lid:binder$ >>,
          <:expr< $inj$ $lid:binder$ >>,
          <:expr< $proj$ a.($`int:index$) >>
      | hd :: tl ->
          let (_loc, inj, proj) = get_ty open_tbl current_tbl hd in
          let binder = new_binder () in
          let inj_pat, inj_arr, proj_tup =
            mk_tup open_tbl current_tbl (succ index) tl in
          Ast.PaCom (_loc, <:patt< $lid:binder$ >>, inj_pat),
          Ast.ExSem (_loc, <:expr< $inj$ $lid:binder$ >>, inj_arr),
          Ast.ExCom (_loc, <:expr< $proj$ a.($`int:index$) >>, proj_tup)

  let translate_arg open_tbl current_tbl ty =
    match ty with
      | Ast.TyId (_loc, Ast.IdLid (_, "unit")) ->
          None, <:expr< () >>
      | _ ->
          let binder = new_binder () in
          let _loc, _, proj = get_ty open_tbl current_tbl ty in
          Some (binder, proj), <:expr< $lid:binder$ >>

  let rec translate_args open_tbl current_tbl ty =
    match ty with
      | Ast.TyArr (_loc, arg, ty) ->
          let lcc_arg, caml_arg = translate_arg open_tbl current_tbl arg in
          let lcc_args_list, caml_args_list, result, decl =
            translate_args open_tbl current_tbl ty in
          let lcc_args_list =
            match lcc_arg with
              | None -> lcc_args_list
              | Some arg -> arg :: lcc_args_list in
          lcc_args_list, caml_arg :: caml_args_list, result, decl
      | _ ->
          let result, decl =
            match ty with
              | Ast.TyId (_loc, Ast.IdLid (_, "unit")) ->
                  let result result_index call =
                    <:expr<
                      $call$;
                      ()
                    >> in
                  result, "decl_do_builtin"
              | Ast.TyId (_loc, Ast.IdLid (_, "bool")) ->
                  let result result_index call = <:expr< $call$ >> in
                  result, "decl_bool_builtin"
              | ty ->
                  let _loc, inj, _ = get_ty open_tbl current_tbl ty in
                  let result result_index call = <:expr< $inj$ $call$ >> in
                  result, "decl_value_builtin" in
          [], [], result, decl

  let translate_value_declaration
      _loc prefix ident open_tbl current_tbl name ty =
    let prefix_name = prefix ^ name in
    try
      counter := 0;
      let lcc_args_list, caml_args_list, result, decl =
        translate_args open_tbl current_tbl ty in
      let add_caml_arg f arg = <:expr< $f$ $arg$ >> in
      let name_lid = Ast.IdLid (_loc, name) in
      let full_name = Ast.ExId (_loc, mk_full_name _loc ident name_lid) in
      let call =
        List.fold_left add_caml_arg full_name caml_args_list in
      let add_lcc_arg (intf, index) (binder, conv) =
        let intf =
          <:expr<
            let $lid:binder$ = $conv$ t.($`int:index$) in
            $intf$
          >> in
        (intf, succ index) in
      let result_index = List.length lcc_args_list in
      let (intf, _) =
        List.fold_left add_lcc_arg
          (result result_index call, 0) lcc_args_list in
      let arity = result_index + 1 in
      <:str_item<
        let () =
          let intf event t =
            $intf$ in
          Interpret.$lid:decl$
            (Kernel.Sgn.get $str:prefix_name$ $`int:arity$)
            intf
      >>
    with Failure msg ->
      Format.eprintf "%s: %s\n" prefix_name msg;
      <:str_item< >>

  let rec translate_sig_declaration prefix ident open_tbl current_tbl sg =
    match sg with
      | Ast.SgTyp (_loc, ty) ->
          translate_type_declaration prefix ident current_tbl ty
      | Ast.SgVal (_loc, name, ty)
      | Ast.SgExt (_loc, name, ty, _) ->
          translate_value_declaration _loc prefix ident open_tbl
            current_tbl name ty
      | Ast.SgSem (_loc, sg0, sg1) ->
          let t0 =
            translate_sig_declaration prefix ident open_tbl current_tbl sg0 in
          let t1 =
            translate_sig_declaration prefix ident open_tbl current_tbl sg1 in
          <:str_item< $t0$ $t1$ >>
      | Ast.SgMod (_loc, m, Ast.MtSig (_, sg)) ->
          let prefix = Format.sprintf "%s%s_" prefix (String.lowercase m) in
          let ident =
            let uid = Ast.IdUid (_loc, m) in
            match ident with
              | None -> uid
              | Some ident -> Ast.IdAcc (_loc, ident, uid) in
          let sub_table = create_module_table () in
          let items =
            translate_sig_declaration prefix (Some ident)
              (current_tbl :: open_tbl) sub_table sg in
          Hashtbl.add current_tbl.modules m sub_table;
          items
      | Ast.SgNil _loc
      | Ast.SgCls (_loc, _)
      | Ast.SgClt (_loc, _)
      | Ast.SgDir (_loc, _, _)
      | Ast.SgExc (_loc, _)
      | Ast.SgInc (_loc, _)
      | Ast.SgMod (_loc, _, _)
      | Ast.SgRecMod (_loc, _)
      | Ast.SgMty (_loc, _, _)
      | Ast.SgOpn (_loc, _)
      | Ast.SgAnt (_loc, _) ->
          <:str_item< >>

  EXTEND Gram
    str_item: [[
      "builtins"; sg = sig_items; "end" ->
        let table = create_module_table () in
        translate_sig_declaration "" None [pervasives_table] table sg
    ]];
  END
end

module M = Camlp4.Register.OCamlSyntaxExtension(Id)(Pa_silcc_builtins)

